"""Base class for Plasma LED devices."""
import atexit


class Plasma():
    """Base class for Plasma LED devices."""

    name = ""

    options = {
        'pixel_count': int,
    }

    option_order = []

    def __init__(self, pixel_count=1):
        """Initialise Plasma device.

        :param pixel_count: Number of individual RGB LEDs

        """
        self._pixel_count = pixel_count
        self._pixels = [[0, 0, 0, 1.0]] * self._pixel_count
        self._clear_on_exit = False

        atexit.register(self.atexit)

    @classmethod
    def parse_options(self, options):
        """Parse option string into kwargs dict."""
        named_options = {}
        for index, option in enumerate(options):
            if "=" in option:
                k, v = option.split("=")
            else:
                v = option
                k = self.option_order[index]
            named_options[k] = self.options[k](v)

        return named_options

    def get_pixel_count(self):
        """Get the count of pixels."""
        return self._pixel_count

    def show(self):
        """Display changes."""
        raise NotImplementedError

    def atexit(self):
        """Clear the display upon exit."""
        if not self._clear_on_exit:
            return
        self.clear()
        self.show()

    def set_pixel_hsv(self, index, h, s, v):
        """Set the HSV colour of an individual pixel in your chain."""
        raise NotImplementedError

    def set_clear_on_exit(self, status=True):
        """Set if the pixel strip should be cleared upon program exit."""
        self._clear_on_exit = status

    def set_all(self, r, g, b, brightness=None):
        """Set the RGB value and optionally brightness of all pixels.

        If you don't supply a brightness value, the last value set for each pixel be kept.

        :param r: Amount of red: 0 to 255
        :param g: Amount of green: 0 to 255
        :param b: Amount of blue: 0 to 255
        :param brightness: Brightness: 0.0 to 1.0 (default is 1.0)

        """
        for x in range(self._pixel_count):
            self.set_pixel(x, r, g, b, brightness)

    def set_sequence(self, sequence):
        """Set RGB values from a Plasma FX sequence."""
        if type(sequence) is list:
            for index, led in enumerate(sequence):
                self.set_pixel(index, *led)
        elif type(sequence) is dict:
            for index, led in sequence.items():
                self.set_pixel(index, *led)
        else:
            for index, led in sequence:
                self.set_pixel(index, *led)

    def get_pixel(self, x):
        """Get the RGB and brightness value of a specific pixel.

        :param x: The horizontal position of the pixel: 0 to 7

        """
        r, g, b, brightness = self._pixels[x]

        return r, g, b, brightness

    def set_pixel(self, x, r, g, b, brightness=None):
        """Set the RGB value, and optionally brightness, of a single pixel.

        If you don't supply a brightness value, the last value will be kept.

        :param x: The horizontal position of the pixel: 0 to 7
        :param r: Amount of red: 0 to 255
        :param g: Amount of green: 0 to 255
        :param b: Amount of blue: 0 to 255
        :param brightness: Brightness: 0.0 to 1.0 (default around 0.2)

        """
        r, g, b = [int(c) & 0xff for c in (r, g, b)]

        if brightness is None:
            brightness = self._pixels[x][3]

        self._pixels[x] = [r, g, b, brightness]

    def clear(self):
        """Clear the pixel buffer."""
        for x in range(self._pixel_count):
            self._pixels[x][0:3] = [0, 0, 0]

    def set_brightness(self, brightness):
        """Set the brightness of all pixels.

        :param brightness: Brightness: 0.0 to 1.0

        """
        if brightness < 0 or brightness > 1:
            raise ValueError('Brightness should be between 0.0 and 1.0')

        for x in range(self._pixel_count):
            self._pixels[x][3] = brightness



"""Class for Plasma light devices in the WS281X/SK6812 family."""
from .core import Plasma


class PlasmaWS281X(Plasma):
    """Class for Plasma light devices in the WS281X/SK6812 family."""

    name = "WS281X"

    options = {
        'pixel_count': int,
        "gpio_pin": int,
        "strip_type": str,
        "channel": int,
        "brightness": int,
        "freq_hz": int,
        "dma": int,
        "invert": bool
    }

    option_order = ("gpio_pin", "strip_type", "channel", "brightness", "freq_hz", "dma", "invert")

    def __init__(self, pixel_count=1, gpio_pin=13, strip_type='WS2812', channel=None, brightness=255, freq_hz=800000, dma=10, invert=False):
        """Initialise WS281X device.

        :param pixel_count: Number of individual RGB LEDs
        :param gpio_pin: BCM GPIO pin for output signal
        :param strip_type: Strip type: one of WS2812 or SK6812
        :param channel: LED channel (0 or 1, or None for automatic)
        :param brightness: Global WS281X LED brightness scale
        :param freq_hz: WS281X output signal frequency (usually 800khz)
        :param dma: DMA channel
        :param invert: Invert signals for NPN-transistor based level shifters

        """
        from rpi_ws281x import PixelStrip, ws

        strip_types = {}
        for t in ws.__dict__:
            if '_STRIP' in t:
                k = t.replace('_STRIP', '')
                v = getattr(ws, t)
                strip_types[k] = v

        strip_type = strip_types[strip_type]

        if channel is None:
            if gpio_pin in [13]:
                channel = 1
            elif gpio_pin in [12, 18]:
                channel = 0

        self._strip = PixelStrip(pixel_count, gpio_pin, freq_hz, dma, invert, brightness, channel, strip_type)
        self._strip.begin()

        Plasma.__init__(self, pixel_count)

    def show(self):
        """Output the buffer."""
        for i in range(self._strip.numPixels()):
            r, g, b, brightness = self._pixels[i]
            self._strip.setPixelColorRGB(i, r, g, b)

        self._strip.show()



"""Plasma support for APA102 style pixels."""
import time
from .core import Plasma


class PlasmaAPA102(Plasma):
    """Plasma support for APA102 style pixels."""

    name = "APA102"

    options = {
        'pixel_count': int,
        "gpio_data": int,
        "gpio_clock": int,
        "gpio_cs": int
    }

    option_order = ("gpio_data", "gpio_clock", "gpio_cs")

    def __init__(self, pixel_count=1, gpio_data=14, gpio_clock=15, gpio_cs=None, gpio=None):
        """Initialise an APA102 device.

        :param pixel_count: Number of individual RGB LEDs
        :param gpio_data: BCM pin for data
        :param gpio_clock: BCM pin for clock
        :param gpio: Optional GPIO back-end, should be RPi.GPIO compatible

        """
        self._gpio = gpio
        if self._gpio is None:
            import RPi.GPIO as GPIO
            self._gpio = GPIO

        self._gpio_data = gpio_data
        self._gpio_clock = gpio_clock
        self._gpio_cs = gpio_cs
        self._gpio_is_setup = False
        Plasma.__init__(self, pixel_count)

    def _write_byte(self, byte):
        for x in range(8):
            self._gpio.output(self._gpio_data, byte & 0b10000000)
            self._gpio.output(self._gpio_clock, 1)
            time.sleep(0)
            byte <<= 1
            self._gpio.output(self._gpio_clock, 0)
            time.sleep(0)

    def _eof(self):
        # Emit exactly enough clock pulses to latch the small dark die APA102s which are weird
        # for some reason it takes 36 clocks, the other IC takes just 4 (number of pixels/2)
        self._gpio.output(self._gpio_data, 0)
        for x in range(36):
            self._gpio.output(self._gpio_clock, 1)
            time.sleep(0)
            self._gpio.output(self._gpio_clock, 0)
            time.sleep(0)

    def _sof(self):
        self._gpio.output(self._gpio_data, 0)
        for x in range(32):
            self._gpio.output(self._gpio_clock, 1)
            time.sleep(0.)
            self._gpio.output(self._gpio_clock, 0)
            time.sleep(0)

    def show(self):
        """Output the buffer."""
        if not self._gpio_is_setup:
            self._gpio.setmode(self._gpio.BCM)
            self._gpio.setwarnings(False)
            self._gpio.setup(self._gpio_data, self._gpio.OUT)
            self._gpio.setup(self._gpio_clock, self._gpio.OUT)
            if self._gpio_cs is not None:
                self._gpio.setup(self._gpio_cs, self._gpio.OUT)

            self._gpio_is_setup = True

        if self._gpio_cs is not None:
            self._gpio.output(self._gpio_cs, 0)

        self._sof()

        for pixel in self._pixels:
            r, g, b, brightness = pixel
            brightness = int(brightness * 31) & 0x1f
            self._write_byte(0b11100000 | brightness)
            self._write_byte(b)
            self._write_byte(g)
            self._write_byte(r)

        self._eof()

        if self._gpio_cs is not None:
            self._gpio.output(self._gpio_cs, 1)



"""Plasma multi device LED driver."""
import sys
import pathlib

__version__ = '2.0.2'


def auto(default=None, descriptor=None):
    """Return a Plasma device instance.

    Will try to get arguments from command-line,
    otherwise falling back to supplied defaults.

    """
    if descriptor is None:
        if len(sys.argv) > 1:
            descriptor = sys.argv[1]
        elif default is not None:
            descriptor = default
        else:
            raise ValueError("get_device requires a descriptor")

    plasma, options = get_device(descriptor)

    return plasma(**options)


def get_device(descriptor=None):
    """Return a Plasma device class and arguments.

    :param descriptor: String describing device and arguments.

    This function accepts descriptors in the form:

    DEVICE:arg:arg

    IE: A serial connection would be described as:

    SERIAL:/dev/ttyACM0

    And GPIO as:

    GPIO:14:15

    And WS281X pixels as:

    WS281X:WS2812_RGB:13:1

    Additional optional arguments can be supplied by name, eg:

    WS281X:WS2812_RGB:13:1:freq_hz=800000

    """
    path = pathlib.Path(descriptor)
    if path.is_file():
        from .matrix import PlasmaMatrix
        return PlasmaMatrix, {'config_file': path}

    if type(descriptor) is str:
        dsc = descriptor.split(":")

        output_type = dsc[0]
        output_options = dsc[1:]

        if output_type in ["GPIO", "APA102"]:
            from .apa102 import PlasmaAPA102
            return PlasmaAPA102, PlasmaAPA102.parse_options(output_options)
        if output_type in ["USB", "SERIAL"]:
            from .serial import PlasmaSerial
            return PlasmaSerial, PlasmaSerial.parse_options(output_options)
        if output_type == "WS281X":
            from .ws281x import PlasmaWS281X
            return PlasmaWS281X, PlasmaWS281X.parse_options(output_options)

    raise ValueError(f"Invalid descriptor: {descriptor}")



"""Wrapper around PlasmaSerial to provide deprecated PlasmaUSB output device."""
from .serial import PlasmaSerial


class PlasmaUSB(PlasmaSerial):
    """Deprecated PlasmaUSB output device."""

    pass



"""Library for the Pimoroni Blinkt! - 8-pixel APA102 LED display."""
import atexit
import time

import RPi.GPIO as GPIO


__version__ = '0.0.1'

DAT = 15
CLK = 14
PIXELS_PER_LIGHT = 4
DEFAULT_BRIGHTNESS = 3
MAX_BRIGHTNESS = 3
NUM_PIXELS = 0


pixels = []

_light_count = 0
_gpio_setup = False
_clear_on_exit = False


def _exit():
    if _clear_on_exit:
        clear()
        show()
    GPIO.cleanup()


def use_pins(data, clock):
    """Set alternate data and clock pins for your Plasma chain."""
    global DAT, CLK, _gpio_setup
    if DAT != data or CLK != clock:
        _gpio_setup = False
    DAT = data
    CLK = clock


def set_light_count(light_count):
    """Set the number of light modules in your Plasma chain."""
    global _light_count, pixels, NUM_PIXELS
    _light_count = light_count
    NUM_PIXELS = light_count * PIXELS_PER_LIGHT
    pixels = [[0, 0, 0, DEFAULT_BRIGHTNESS]] * light_count * PIXELS_PER_LIGHT


def set_light(index, r, g, b):
    """Set the RGB colour of an individual light in your Plasma chain.

    This will set all four LEDs on the Plasma light to the same colour.

    :param index: Index of the light in your chain (starting at 0)
    :param r: Amount of red: 0 to 255
    :param g: Amount of green: 0 to 255
    :param b: Amount of blue: 0 to 255

    """
    offset = index * 4
    for x in range(4):
        set_pixel(offset + x, r, g, b)


def set_brightness(brightness):
    """Set the brightness of all pixels.

    :param brightness: Brightness: 0.0 to 1.0

    """
    if brightness < 0 or brightness > 1:
        raise ValueError('Brightness should be between 0.0 and 1.0')

    for x in range(_light_count * PIXELS_PER_LIGHT):
        pixels[x][3] = int(float(MAX_BRIGHTNESS) * brightness) & 0b11111


def clear():
    """Clear the pixel buffer."""
    for x in range(_light_count * PIXELS_PER_LIGHT):
        pixels[x][0:3] = [0, 0, 0]


def _write_byte(byte):
    for x in range(8):
        GPIO.output(DAT, byte & 0b10000000)
        GPIO.output(CLK, 1)
        time.sleep(0.0000005)
        byte <<= 1
        GPIO.output(CLK, 0)
        time.sleep(0.0000005)


# Emit exactly enough clock pulses to latch the small dark die APA102s which are weird
# for some reason it takes 36 clocks, the other IC takes just 4 (number of pixels/2)
def _eof():
    GPIO.output(DAT, 0)
    for x in range(36):
        GPIO.output(CLK, 1)
        time.sleep(0.0000005)
        GPIO.output(CLK, 0)
        time.sleep(0.0000005)


def _sof():
    GPIO.output(DAT, 0)
    for x in range(32):
        GPIO.output(CLK, 1)
        time.sleep(0.0000005)
        GPIO.output(CLK, 0)
        time.sleep(0.0000005)


def show():
    """Output the buffer to Blinkt!."""
    global _gpio_setup

    if not _gpio_setup:
        GPIO.setmode(GPIO.BCM)
        GPIO.setwarnings(False)
        GPIO.setup(DAT, GPIO.OUT)
        GPIO.setup(CLK, GPIO.OUT)
        atexit.register(_exit)
        _gpio_setup = True

    _sof()

    for pixel in pixels:
        r, g, b, brightness = pixel
        _write_byte(0b11100000 | brightness)
        _write_byte(b)
        _write_byte(g)
        _write_byte(r)

    _eof()


def set_all(r, g, b, brightness=None):
    """Set the RGB value and optionally brightness of all pixels.

    If you don't supply a brightness value, the last value set for each pixel be kept.

    :param r: Amount of red: 0 to 255
    :param g: Amount of green: 0 to 255
    :param b: Amount of blue: 0 to 255
    :param brightness: Brightness: 0.0 to 1.0 (default is 1.0)

    """
    for x in range(_light_count * PIXELS_PER_LIGHT):
        set_pixel(x, r, g, b, brightness)


def get_pixel(x):
    """Get the RGB and brightness value of a specific pixel.

    :param x: The horizontal position of the pixel: 0 to 7

    """
    r, g, b, brightness = pixels[x]
    brightness /= float(MAX_BRIGHTNESS)

    return r, g, b, round(brightness, 3)


def set_pixel(x, r, g, b, brightness=None):
    """Set the RGB value, and optionally brightness, of a single pixel.

    If you don't supply a brightness value, the last value will be kept.

    :param x: The horizontal position of the pixel: 0 to 7
    :param r: Amount of red: 0 to 255
    :param g: Amount of green: 0 to 255
    :param b: Amount of blue: 0 to 255
    :param brightness: Brightness: 0.0 to 1.0 (default around 0.2)

    """
    if brightness is None:
        brightness = pixels[x][3]
    else:
        brightness = int(float(MAX_BRIGHTNESS) * brightness) & 0b11111

    pixels[x] = [int(r) & 0xff, int(g) & 0xff, int(b) & 0xff, brightness]


def set_clear_on_exit(value=True):
    """Set whether Plasma Lights should be cleared upon exit.

    By default Plasma will not turn off the pixels on exit, but calling::

        plasma.set_clear_on_exit(True)

    Will ensure that it does.

    :param value: True or False (default True)

    """
    global _clear_on_exit
    _clear_on_exit = value



"""Serial class for Plasma light devices over USB Serial/UART."""
from .core import Plasma
from serial import Serial


class PlasmaSerial(Plasma):
    """Serial class for Plasma light devices over USB Serial/UART."""

    name = "Serial"

    options = {
        'pixel_count': int,
        "port": str
    }

    option_order = ("port", )

    def __init__(self, pixel_count=1, port='/dev/ttyAMA0', baudrate=115200):
        """Initialise Serial device.

        :param pixel_count: Number of individual RGB LEDs
        :param port: Serial port name/path
        :param baudrate: Serial baud rate

        """
        self._serial_port = port
        self._serial = Serial(port, baudrate=baudrate)
        Plasma.__init__(self, pixel_count)

    def show(self):
        """Display current buffer on LEDs."""
        sof = b"LEDS"
        eof = b"\x00\x00\x00\xff"

        pixels = []
        for pixel in self._pixels:
            r, g, b, brightness = pixel
            pixels += [r, g, b, int(brightness * 254)]

        self._serial.write(sof + bytearray(pixels) + eof)
        self._serial.flush()



"""Combine multiple LED strip types into a single logical strip."""
import pathlib
import yaml


class PlasmaMatrix():
    """Combine multiple LED strip types into a single logical strip."""

    def __init__(self, config_file=None):
        """Initialise a matrix.

        :param config_file: Path to yml configuration file

        """
        self._devices = {}

        if type(config_file) is str:
            config_file = pathlib.Path(config_file)

        if not config_file.is_file():
            raise ValueError(f"Could not find {config_file}")

        self._config = yaml.safe_load(open(config_file, "r"))

        for required in ("pixels", "devices"):
            if required not in self._config:
                raise ValueError(f"Config missing required setting: '{required}'")

        self._pixel_count = int(self._config["pixels"])

        for output_name, output_options in self._config["devices"].items():
            enabled = output_options.get("enabled", True)
            if not enabled:
                continue

            output_type = output_options.get("type")
            pixels = output_options.get("pixels", self._pixel_count)
            offset = output_options.get("offset", 0)

            del output_options["type"]

            try:
                del output_options["enabled"]
            except KeyError:
                pass

            try:
                del output_options["pixels"]
            except KeyError:
                pass

            try:
                del output_options["offset"]
            except KeyError:
                pass

            output_device = self.get_output_device(output_type)

            self._devices[output_name] = {
                'offset': offset,
                'type': output_type,
                'device': output_device(pixels, **output_options)
            }

    def get_pixel_count(self):
        """Get the count of pixels."""
        return self._pixel_count

    def get_device_count(self):
        """Get the count of output devices."""
        return len(self._devices)

    def get_device(self, index):
        """Get a device by name or type.

        If a string (type) is given, it will try to find a device with that name.

        Otherwise returns first device of that type.

        """
        if index in self._devices.keys():
            return self._devices[index].get('device')

        if type(index) == str:
            for n, d in self._devices.items():
                if d.get('type') == index:
                    return d.get('device')

        raise ValueError(f"Invalid device index {index}")

    def get_output_device(self, output_type):
        """Get output class by name."""
        output_type = output_type.upper()
        if output_type in ["GPIO", "APA102"]:
            from .gpio import PlasmaGPIO
            return PlasmaGPIO
        if output_type in ["USB", "SERIAL"]:
            from .usb import PlasmaSerial
            return PlasmaSerial
        if output_type == "WS281X":
            from .ws281x import PlasmaWS281X
            return PlasmaWS281X
        raise ValueError(f"Invalid output type {output_type}!")

    def set_light(self, index, r, g, b, brightness=None):
        """Set the RGB colour of an individual light in your matrix."""
        raise NotImplementedError("Use get_device(index).set_light()")

    def set_all(self, r, g, b, brightness=None):
        """Set the RGB value and optionally brightness of all pixels."""
        for n, d in self._devices.items():
            d.get('device').set_all(r, g, b, brightness=brightness)

    def set_sequence(self, sequence):
        """Set all LEDs from a buffer of individual colours."""
        if type(sequence) is list:
            for index, led in enumerate(sequence):
                self.set_pixel(index, *led)
        elif type(sequence) is dict:
            for index, led in sequence.items():
                self.set_pixel(index, *led)
        else:
            for index, led in sequence:
                self.set_pixel(index, *led)

    def get_pixel(self, x):
        """Get the RGB and brightness value of a specific pixel."""
        for n, d in self._devices.items():
            device = d.get('device')
            offset = d.get('offset')
            count = device.get_pixel_count()

            if x >= offset and x < offset + count:
                return device.get_pixel(x - offset)

    def set_pixel(self, x, r, g, b, brightness=None):
        """Set the RGB value, and optionally brightness, of a single pixel."""
        for n, d in self._devices.items():
            device = d.get('device')
            offset = d.get('offset')
            count = device.get_pixel_count()

            if x >= offset and x < offset + count:
                device.set_pixel(x - offset, r, g, b, brightness=brightness)

    def show(self):
        """Display lights across devices.

        Sets each device in turn, picking the pixels
        from the buffer starting from the specified offset.

        """
        for n, d in self._devices.items():
            d.get('device').show()
